from random import shuffle
from django.conf import settings
try: # python >=3.8
    from importlib.metadata import version as get_version
except ModuleNotFoundError: # python < 3.8
    from pkg_resources import get_distribution as get_version

INFSCROLL_VERSION = get_version('django-infinite-scroll')
PAGINATION_STEPS = int(getattr(settings, 'PAGINATION_STEPS', 10))

def get_pagination(request, feed:list,
                   pagination_steps:int=PAGINATION_STEPS,
                   minimum:int=0,
                   page_canonical:int=None,
                   shuf:bool=False,
                   forever:bool=True) -> dict:
    """ Prepare data to send html in order to keep a syncronized pagination.
    :param request HttpRequest: the request
    :param pagination_steps int: how many items will load for each page.
    :param minimum int: the minimum amount of items in the list to trigger
    :param page_canonical int: the absolute page number.
                                eg.Page 2 is from 20 to 30.
    :param shuf bool: if we should shuffle the array.
    :param forever bool: If this is true, when we get at the end of the list,
                         start from the begining.
    """
    has_newer_posts = True
    has_older_posts = True
    if 'view-all' in request.GET.keys():
        pagination_steps=len(feed)
    if page_canonical:
        current_page_canonical = int(page_canonical) * pagination_steps
        page = int(current_page_canonical)
    else:
        page = num_only(request.GET.get('page', 0))
        page_canonical = int(page / pagination_steps)
    older_posts = page + pagination_steps
    newer_posts = page - pagination_steps
    if page <= 0:
        newer_posts = 0
        page = 0
        has_newer_posts = False
    feed_count = len(feed)
    feed = feed[page:older_posts]
    should_load_more = (feed_count >= minimum)
    # This means we reached the end of the list
    if feed_count <= older_posts:
        has_older_posts = False
        older_posts = 0
        should_load_more = (should_load_more and forever)
    if shuf:
        feed = list(feed)
        shuffle(feed)
    data = {
        'feed': feed,
        'page': page,
        'older_posts': older_posts,
        'newer_posts': newer_posts,
        'has_newer_posts': has_newer_posts,
        'has_older_posts': has_older_posts,
        'should_load_more': should_load_more,
        'page_canonical': page_canonical,
        'older_posts_canonica': int(older_posts / pagination_steps),
        'newer_posts_canonica': int(newer_posts / pagination_steps),
        'infscroll_version': INFSCROLL_VERSION,
        }
    return data

def num_only(value:any)->int:
    """ extract non digits from a string and return integer """
    value = str(value)
    nlist = [num for num in value if num.isdigit()]
    if nlist:
        return int(''.join(nlist))
    else:
        return None




from django import template
from django.template.loader import render_to_string

register = template.Library()

@register.simple_tag(name='set_infinite_scroll', takes_context=True)
def set_infinite_scroll(context):
    """ renders the html to set up and load the necessary JS and CSS.
    This is used by {% set_infinite_scroll %} tag.
    """
    data = get_context_data(context)
    return render_to_string('infscroll/scroll.html', data)

@register.simple_tag(name='infinite_scroll_box', takes_context=True)
def infinite_scroll_box(context):
    """ renders the html that loads the dynamic content.
    This is used by {% infinite_scroll_box %} tag.
    """
    data = get_context_data(context)
    return render_to_string('infscroll/scroll_box.html', data)

@register.simple_tag(name='infinite_scroll_tags', takes_context=True)
def infinite_scroll_tags(context):
    """ renders the html that loads the dynamic content.
    This is used by {% infinite_scroll_box %} tag.
    """
    data = get_context_data(context)
    return render_to_string('infscroll/scroll_tags.html', data)


def get_context_data(context):
    """ Get the correct dictionary from the context.
    Maybe there is a better way to do this...
    """
    for ctx_dict in context.dicts:
        if 'feed' in ctx_dict.keys():
            return ctx_dict



from django.shortcuts import render
from infscroll.utils import get_pagination, PAGINATION_STEPS

def more_items(request, feed:list,
               template:str='infscroll/more.html',
               extra_data:dict={},
               steps:int=PAGINATION_STEPS,
               minimum:int=0,
               shuf:bool=False,
               forever:bool=True):
    """ 
    Renders a page with the next round of items in pagination. 
    To be used with ajax calls
    """
    data = get_pagination(request, feed, pagination_steps=steps,
                          minimum=minimum, shuf=shuf, forever=forever)
    data.update(extra_data)
    return render(request, template, data)



from django.test import TestCase

from infscroll.utils import get_pagination, num_only
from infscroll.views import more_items

class FakeRequest(object):
    def __init__(self, get={}):
        self.GET = get

# Create your tests here.
class InfiniteScrollTestCase(TestCase):
    def setUp(self):
        self.feed = list(range(100))

    def test_get_pagination(self):
        """ test the initial pagination step """
        request = FakeRequest()
        data = get_pagination(request, self.feed,
                              pagination_steps=10)
        self.assertEqual(len(data['feed']), 10)
        i = 0
        for item in data['feed']:
            assert self.feed[i] == item
            i += 1
        self.assertEqual(data['page'], 0)
        self.assertEqual(data['older_posts'], 10)
        self.assertEqual(data['newer_posts'], 0)
        self.assertEqual(data['has_newer_posts'], False)
        self.assertEqual(data['has_older_posts'], True)
        self.assertEqual(data['should_load_more'], True)
        self.assertEqual(data['page_canonical'], 0)
        self.assertEqual(data['older_posts_canonica'], 1)
        self.assertEqual(data['newer_posts_canonica'], 0)

    def test_get_pagination_goes_up(self):
        """ test what would be the second loaded page """
        request = FakeRequest({'page': '20'})
        data = get_pagination(request, self.feed,
                              pagination_steps=10)
        self.assertEqual(len(data['feed']), 10)
        i = 20
        for item in data['feed']:
            assert self.feed[i] == item
            i += 1
        self.assertEqual(data['page'], 20)
        self.assertEqual(data['older_posts'], 30)
        self.assertEqual(data['newer_posts'], 10)
        self.assertEqual(data['has_newer_posts'], True)
        self.assertEqual(data['has_older_posts'], True)
        self.assertEqual(data['should_load_more'], True)
        self.assertEqual(data['page_canonical'], 2)
        self.assertEqual(data['older_posts_canonica'], 3)
        self.assertEqual(data['newer_posts_canonica'], 1)

    def test_get_pagination_canonica(self):
        """ check the pagination when the canonical number is called
        Eg. Page 2 should load from 20 to 30
        """
        request = FakeRequest()
        data = get_pagination(request, self.feed,
                              page_canonical=2,
                              pagination_steps=10)
        self.assertEqual(len(data['feed']), 10)
        i = 20
        for item in data['feed']:
            assert self.feed[i] == item
            i += 1
        self.assertEqual(data['page'], 20)
        self.assertEqual(data['older_posts'], 30)
        self.assertEqual(data['newer_posts'], 10)
        self.assertEqual(data['has_newer_posts'], True)
        self.assertEqual(data['has_older_posts'], True)
        self.assertEqual(data['should_load_more'], True)
        self.assertEqual(data['page_canonical'], 2)
        self.assertEqual(data['older_posts_canonica'], 3)
        self.assertEqual(data['newer_posts_canonica'], 1)

    def test_get_pagination_end_of_list(self):
        """ Test the behavior when we get to the end of our list.
        Eg. From a list of 100 items, the end is 90 to 99
        """
        request = FakeRequest({'page': '90'})
        data = get_pagination(request, self.feed,
                              pagination_steps=10)
        self.assertEqual(len(data['feed']), 10)
        i = 90
        for item in data['feed']:
            assert self.feed[i] == item
            i += 1
        self.assertEqual(data['page'], 90)
        self.assertEqual(data['older_posts'], 0)
        self.assertEqual(data['newer_posts'], 80)
        self.assertEqual(data['has_newer_posts'], True)
        self.assertEqual(data['has_older_posts'], False)
        self.assertEqual(data['should_load_more'], True)
        self.assertEqual(data['page_canonical'], 9)
        self.assertEqual(data['older_posts_canonica'], 0)
        self.assertEqual(data['newer_posts_canonica'], 8)

    def test_get_pagination_off_the_list(self):
        """ Test when we end our list but there is less than a page of items.
        Eg. Our list has 100 items, If we start at 95 it should be from 95 to 99
        """
        request = FakeRequest({'page': '95'})
        data = get_pagination(request, self.feed,
                              pagination_steps=10)
        self.assertEqual(len(data['feed']), 5)
        i = 95
        for item in data['feed']:
            assert self.feed[i] == item
            i += 1
        self.assertEqual(data['page'], 95)
        self.assertEqual(data['older_posts'], 0)
        self.assertEqual(data['newer_posts'], 85)
        self.assertEqual(data['has_newer_posts'], True)
        self.assertEqual(data['has_older_posts'], False)
        self.assertEqual(data['should_load_more'], True)
        self.assertEqual(data['page_canonical'], 9)
        self.assertEqual(data['older_posts_canonica'], 0)
        self.assertEqual(data['newer_posts_canonica'], 8)

    def test_get_pagination_view_all(self):
        """ Test when the ?view-all flag is set. Should disable pagination """
        request = FakeRequest({'view-all': None})
        data = get_pagination(request, self.feed,
                              pagination_steps=10)
        self.assertEqual(len(data['feed']), len(self.feed))
        i = 0
        for item in data['feed']:
            assert self.feed[i] == item
            i += 1

    def test_get_pagination_shuffle(self):
        """ test when we ask to shuffle the array """
        request = FakeRequest()
        data = get_pagination(request, self.feed,
                              pagination_steps=10,
                              shuf=True)
        self.assertEqual(len(data['feed']), 10)
        for item in data['feed']:
            assert item in self.feed[0:10]

    def test_not_forever(self):
        """ Test when we get to the end of our list, and the flag *forever* is
        set to false.
        """
        request = FakeRequest({'page': '90'})
        data = get_pagination(request, self.feed,
                              pagination_steps=10,
                              forever=False)
        self.assertEqual(len(data['feed']), 10)
        self.assertEqual(data['page'], 90)
        self.assertEqual(data['older_posts'], 0)
        self.assertEqual(data['newer_posts'], 80)
        self.assertEqual(data['has_newer_posts'], True)
        self.assertEqual(data['has_older_posts'], False)
        self.assertEqual(data['should_load_more'], False)

    def test_view_more_items(self):
        """ test the view that loads more items """
        request = FakeRequest()
        response = more_items(request, self.feed)
        self.assertEqual(response.status_code, 200)
        data = get_pagination(request, self.feed,
                              pagination_steps=10)
        content = str(response.content)
        for item in data['feed']:
            self.assertIn(str(item), content)

    def test_num_only(self):
        """ unitary test for the function num_only """
        assert num_only(35) == 35
        assert num_only("35") == 35
        assert num_only("3.5") == 35
        assert num_only("3a$d5") == 35
        assert num_only(0) == 0
        assert num_only("0") == 0
        assert num_only("") == None
        assert num_only(" ") == None
